==> bibledit-0.7.tar.gz: bibledit-0.7/src/worddatabase.cpp <==
/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "worddatabase.h"
#include <config.h>
#include "sqlite3_plus.h"
#include "usfmtools.h"
#include "bible.h"
#include "log.h"
#include "book.h"
#include "usfm.h"
#include "scripture.h"
#include <set>
#include <glib.h>


#define TABLE_FILE_SIGNATURE "file_signature"
#define TABLE_LINES "lines"


Worddatabase::Worddatabase (const string & project)
{
  // Initialize variables.
  myproject = project;
}


Worddatabase::~Worddatabase ()
{
}


void
Worddatabase::initialize ()
{
  try
  {
    // Connect to the database and create the table(s) if needed.
    sqlite3::connection con (dbfilename ().c_str ());
    int count;
    // Check for the table that stores the lines
    count = con.executeint32("select count(*) from sqlite_master where name='%s';", TABLE_LINES);
    if (count == 0)
      con.executenonquery("create table '%s' (book integer, chapter integer, verse integer, line text, casefolded text);", TABLE_LINES);
    // Check for the table that stores the file signature
    count = con.executeint32 ("select count(*) from sqlite_master where name='%s';",  TABLE_FILE_SIGNATURE);
    if (count == 0)
      con.executenonquery ("create table %s (filename text, signature text);", TABLE_FILE_SIGNATURE);
    con.close ();
    // Get books in project.
    Scripture scripture (myproject);
    // Delete any books not in the project.
    set <string> currentbooks;
    for (unsigned int i = 0; i < scripture.books.size (); i++) {
      currentbooks.insert (scripture.books[i]);
    }
    for (unsigned int i = 0; i < NUMBER_OF_BIBLEBOOKS; i++) {
      if (currentbooks.find (index_to_biblebook (i)) == currentbooks.end ()) {
        delete_book (i);
      }
    }
    // Check each book.
    for (unsigned int i = 0; i < scripture.paths.size (); i++)
      check_book (scripture.paths[i]);
  }
  catch (exception & ex)
  {
    cerr << "Error initializing: " << ex.what () << endl;
  }
}


void
Worddatabase::update_chapter (unsigned int book, unsigned int chapter, const vector <ustring>& lines)
{
  log ("Updating " + index_to_biblebook (book) + " " + convert_to_string (chapter));
  try
  {
    /* 
      In order to avoid getting several times the same verse, when searching,
      we have to ensure only one line per verse is stored in the database.
      We first store all data in memory, and at the end add it to the database.
    */
    // Buffers where all data to be stored is collected.
    vector<unsigned int> book_to_store;
    vector<unsigned int> chapter_to_store;
    vector<string> verse_to_store;
    vector<ustring> line_to_store;
    vector<ustring> casefolded_line_to_store;
    string previous_verse;
    // Process each line.
    string currentverse = "1";
    Usfm usfm (0);
    for (unsigned int i = 0; i < lines.size (); i++) {
      ustring line (lines[i]);
      // Normalize the string for better result when comparing text.
      line = line.normalize ();
      // Extract the verse number.
      ustring string_line;
      string_line = line.raw ();
      ustring marker;
      marker = usfm_extract_marker (string_line);
      if (usfm.is_verse (marker))
        currentverse = number_in_string (string_line);
      // Any apostrophe (') should be prefixed by another one.
      line = double_apostrophy (line);
      ustring casefolded (line);
      casefolded = casefolded.casefold ();
      // Store the data in the buffers.
      if (currentverse != previous_verse) {
        // We've a new verse, so add it to the end of the buffers.
        previous_verse = currentverse;
        book_to_store.push_back(book);
        chapter_to_store.push_back(chapter);
        verse_to_store.push_back(currentverse);
        line_to_store.push_back(line);
        casefolded_line_to_store.push_back(casefolded);
      } else {
        // We've the same verse, so add it to the last items in the buffers.
        unsigned int last_item = line_to_store.size() - 1;
        line_to_store[last_item].append (" " + line);
        casefolded_line_to_store[last_item].append (" " + casefolded);
      }
    }
    // Connect and start transaction
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    // Delete previous data (if any);
    con.executenonquery ("delete from '%s' where book = %d and chapter = %d;", TABLE_LINES, book, chapter);
    // Put data from the buffers in the database.
    for (unsigned int i = 0; i < line_to_store.size(); i++) {
      con.executenonquery ("insert into '%s' values (%d, %d, %s, '%s', '%s');", TABLE_LINES, book_to_store[i], chapter_to_store[i], verse_to_store[i].c_str(), line_to_store[i].c_str(), casefolded_line_to_store[i].c_str());
    }
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error updating: " << ex.what () << endl;
  }
}


void
Worddatabase::log (const string & message)
{
  log_message ("Words database: " + message);
}


void
Worddatabase::check_book (const string & filename)
{
  try
  {
    // Connect to the database
    sqlite3::connection con (dbfilename ().c_str ());
    // Get the actual signature for the file.
    string actual_signature;
    /* 
     * Before we got the signature through the ls -l command, but this caused 
     * a problem. It gives the modification time of the file with a resolution
     * of a minute. But if changes are applied to a file, and after the change
     * it still shows the same minute, then it did not get updates in the words
     * database. We now use the stat system call with higher accuracy of time.
     * This solves the problem 
     */
    struct stat statbuf;
    stat (filename.c_str(), &statbuf);
    actual_signature = convert_to_string (int(statbuf.st_size));
    actual_signature.append (convert_to_string (int(statbuf.st_mtime)));
    actual_signature.append (convert_to_string (int(statbuf.st_ctime)));
    // See if the filename exists in the table anyway.
    bool update = false;
    int count = con.executeint32 ("select count(*) from '%s' where filename='%s';",
                                  TABLE_FILE_SIGNATURE, filename.c_str ());
    if (count == 0)
      {
        // Doesn't exists, so schedule the file for updating the database.
        update = true;
      }
    if (!update)
      {
        // The filename is in the database - now check its signature.
        sqlite3::reader reader = con.executereader ("select signature from '%s' where filename='%s';",
                                                    TABLE_FILE_SIGNATURE, filename.c_str ());
        if (reader.read ())
          {
            string stored_signature = reader.getstring (0);
            if (stored_signature != actual_signature)
              {
                update = true;
              }
          }
        reader.close ();
      }
    if (update)
      {
        log ("Will be updated by " + filename);
        // Store data about this filename.
        Book wbook (filename, false, false);
        for (unsigned int chapter = 1; chapter <= wbook.number_of_chapters;
             chapter++)
          {
            vector <ustring> lines;
            wbook.get_chapter (chapter, lines);
            update_chapter (wbook.get_id_index (), chapter, lines);
          }
        // Store its signature.
        store_signature (filename, actual_signature);
      }
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error checking: " << ex.what () << endl;
  }
}


void
Worddatabase::delete_book (unsigned int book)
{
  try
  {
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where book = %d;", TABLE_LINES, book);
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error deleting: " << ex.what () << endl;
  }
}


string
Worddatabase::dbfilename ()
{
  // Returns full path and name of the word database.
  string s = myproject + "_wordsdatabase.sqlite3";
  return g_build_filename (PACKAGE_LOCALSTATE_DIR, s.c_str(), NULL);
}


void
Worddatabase::store_signature (const string & filename,
                               const string & signature)
{
  // Store the signature of the book.
  try
  {
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where filename = '%s';",
                         TABLE_FILE_SIGNATURE, filename.c_str ());
    con.executenonquery ("insert into '%s' values('%s', '%s');",
                         TABLE_FILE_SIGNATURE, filename.c_str (),
                         signature.c_str ());
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error signing: " << ex.what () << endl;
  }
}

==> bibledit-0.9.tar.gz: bibledit-0.9/src/worddatabase.cpp <==
/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "worddatabase.h"
#include <config.h>
#include "sqlite3_plus.h"
#include "usfmtools.h"
#include "bible.h"
#include "log.h"
#include "book.h"
#include "usfm.h"
#include "scripture.h"
#include <set>
#include <glib.h>
#include "project.h"


#define TABLE_FILE_SIGNATURE "file_signature"
#define TABLE_LINES "lines"


Worddatabase::Worddatabase (const string & project)
{
  // Initialize variables.
  myproject = project;
  Project prj (project);
  usfm = new Usfm (prj.stylesheet);
}


Worddatabase::~Worddatabase ()
{
  delete usfm;
}


void
Worddatabase::initialize ()
{
  try
  {
    // Connect to the database and create the table(s) if needed.
    sqlite3::connection con (dbfilename ().c_str ());
    int count;
    // Check for the table that stores the lines
    count = con.executeint32("select count(*) from sqlite_master where name='%s';", TABLE_LINES);
    if (count == 0)
      con.executenonquery("create table '%s' (book integer, chapter integer, verse integer, line text, casefolded text);", TABLE_LINES);
    // Check for the table that stores the file signature
    count = con.executeint32 ("select count(*) from sqlite_master where name='%s';",  TABLE_FILE_SIGNATURE);
    if (count == 0)
      con.executenonquery ("create table %s (filename text, signature text);", TABLE_FILE_SIGNATURE);
    con.close ();
    // Get books in project.
    Scripture scripture (myproject);
    // Delete any books not in the project.
    set <string> currentbooks;
    for (unsigned int i = 0; i < scripture.books.size (); i++) {
      currentbooks.insert (scripture.books[i]);
    }
    for (unsigned int i = 0; i < NUMBER_OF_BIBLEBOOKS; i++) {
      if (currentbooks.find (index_to_biblebook (i)) == currentbooks.end ()) {
        delete_book (i);
      }
    }
    // Check each book.
    for (unsigned int i = 0; i < scripture.paths.size (); i++)
      check_book (scripture.paths[i]);
  }
  catch (exception & ex)
  {
    cerr << "Error initializing: " << ex.what () << endl;
  }
}


void
Worddatabase::update_chapter (unsigned int book, unsigned int chapter, const vector <ustring>& lines)
{
  log ("Updating " + index_to_biblebook (book) + " " + convert_to_string (chapter));
  try
  {
    /* 
      In order to avoid getting several times the same verse, when searching,
      we have to ensure only one line per verse is stored in the database.
      We first store all data in memory, and at the end add it to the database.
    */
    // Buffers where all data to be stored is collected.
    vector<unsigned int> book_to_store;
    vector<unsigned int> chapter_to_store;
    vector<string> verse_to_store;
    vector<ustring> line_to_store;
    vector<ustring> casefolded_line_to_store;
    string previous_verse;
    // Process each line.
    string currentverse = "1";
    for (unsigned int i = 0; i < lines.size (); i++) {
      ustring line (lines[i]);
      // Normalize the string for better result when comparing text.
      line = line.normalize ();
      // Extract the verse number.
      ustring string_line;
      string_line = line.raw ();
      ustring marker;
      marker = usfm_extract_marker (string_line);
      if (usfm->is_verse_number (marker))
        currentverse = number_in_string (string_line);
      // Any apostrophe (') should be prefixed by another one.
      line = double_apostrophy (line);
      ustring casefolded (line);
      casefolded = casefolded.casefold ();
      // Store the data in the buffers.
      if (currentverse != previous_verse) {
        // We've a new verse, so add it to the end of the buffers.
        previous_verse = currentverse;
        book_to_store.push_back(book);
        chapter_to_store.push_back(chapter);
        verse_to_store.push_back(currentverse);
        line_to_store.push_back(line);
        casefolded_line_to_store.push_back(casefolded);
      } else {
        // We've the same verse, so add it to the last items in the buffers.
        unsigned int last_item = line_to_store.size() - 1;
        line_to_store[last_item].append (" " + line);
        casefolded_line_to_store[last_item].append (" " + casefolded);
      }
    }
    // Connect and start transaction
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    // Delete previous data (if any);
    con.executenonquery ("delete from '%s' where book = %d and chapter = %d;", TABLE_LINES, book, chapter);
    // Put data from the buffers in the database.
    for (unsigned int i = 0; i < line_to_store.size(); i++) {
      con.executenonquery ("insert into '%s' values (%d, %d, %s, '%s', '%s');", TABLE_LINES, book_to_store[i], chapter_to_store[i], verse_to_store[i].c_str(), line_to_store[i].c_str(), casefolded_line_to_store[i].c_str());
    }
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error updating: " << ex.what () << endl;
  }
}


void
Worddatabase::log (const string & message)
{
  log_message ("Words database: " + message);
}


void
Worddatabase::check_book (const string & filename)
{
  try
  {
    // Connect to the database
    sqlite3::connection con (dbfilename ().c_str ());
    // Get the actual signature for the file.
    string actual_signature;
    /* 
     * Before we got the signature through the ls -l command, but this caused 
     * a problem. It gives the modification time of the file with a resolution
     * of a minute. But if changes are applied to a file, and after the change
     * it still shows the same minute, then it did not get updates in the words
     * database. We now use the stat system call with higher accuracy of time.
     * This solves the problem 
     */
    struct stat statbuf;
    stat (filename.c_str(), &statbuf);
    actual_signature = convert_to_string (int(statbuf.st_size));
    actual_signature.append (convert_to_string (int(statbuf.st_mtime)));
    actual_signature.append (convert_to_string (int(statbuf.st_ctime)));
    // See if the filename exists in the table anyway.
    bool update = false;
    int count = con.executeint32 ("select count(*) from '%s' where filename='%s';",
                                  TABLE_FILE_SIGNATURE, filename.c_str ());
    if (count == 0)
      {
        // Doesn't exists, so schedule the file for updating the database.
        update = true;
      }
    if (!update)
      {
        // The filename is in the database - now check its signature.
        sqlite3::reader reader = con.executereader ("select signature from '%s' where filename='%s';",
                                                    TABLE_FILE_SIGNATURE, filename.c_str ());
        if (reader.read ())
          {
            string stored_signature = reader.getstring (0);
            if (stored_signature != actual_signature)
              {
                update = true;
              }
          }
        reader.close ();
      }
    if (update)
      {
        log ("Will be updated by " + filename);
        // Store data about this filename.
        Book wbook (filename, false, false, usfm);
        for (unsigned int chapter = 1; chapter <= wbook.number_of_chapters;
             chapter++)
          {
            vector <ustring> lines;
            wbook.get_chapter (chapter, lines);
            update_chapter (wbook.get_id_index (), chapter, lines);
          }
        // Store its signature.
        store_signature (filename, actual_signature);
      }
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error checking: " << ex.what () << endl;
  }
}


void
Worddatabase::delete_book (unsigned int book)
{
  try
  {
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where book = %d;", TABLE_LINES, book);
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error deleting: " << ex.what () << endl;
  }
}


string
Worddatabase::dbfilename ()
{
  // Returns full path and name of the word database.
  string s = myproject + "_wordsdatabase.sqlite3";
  return g_build_filename (PACKAGE_LOCALSTATE_DIR, s.c_str(), NULL);
}


void
Worddatabase::store_signature (const string & filename,
                               const string & signature)
{
  // Store the signature of the book.
  try
  {
    sqlite3::connection con (dbfilename ().c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where filename = '%s';",
                         TABLE_FILE_SIGNATURE, filename.c_str ());
    con.executenonquery ("insert into '%s' values('%s', '%s');",
                         TABLE_FILE_SIGNATURE, filename.c_str (),
                         signature.c_str ());
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error signing: " << ex.what () << endl;
  }
}
